import pytest

from aioecowitt import server


@pytest.fixture
def ecowitt_server():
    """EcoWitt server fixture."""
    ecowitt_server = server.EcoWittListener()
    yield ecowitt_server


@pytest.fixture
def ecowitt_http(event_loop, aiohttp_raw_server, aiohttp_client, ecowitt_server):
    """EcoWitt HTTP fixture."""
    raw_server = event_loop.run_until_complete(
        aiohttp_raw_server(ecowitt_server.handler)
    )
    return event_loop.run_until_complete(aiohttp_client(raw_server))



"""EcoWitt server tests."""

import pytest

from aioecowitt import server

from .const import GW2000A_DATA, EASYWEATHER_DATA

# pylint: disable=redefined-outer-name


@pytest.mark.asyncio
async def test_server_start(ecowitt_server, ecowitt_http) -> None:
    """Test server start."""
    sensors = []

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        sensors.append(sensor)

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert len(sensors) == 52
    assert len(ecowitt_server.sensors) == 52
    assert len(ecowitt_server.stations) == 1

    assert "PASSKEY" not in ecowitt_server.last_values[GW2000A_DATA["PASSKEY"]]


@pytest.mark.asyncio
async def test_server_token(ecowitt_server, ecowitt_http) -> None:
    """Test server start."""
    sensors = []
    path = "/test"
    ecowitt_server.path = path

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        sensors.append(sensor)

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_DATA)
    assert resp.status == 404

    resp = await ecowitt_http.post(path, data=GW2000A_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert len(sensors) == 52
    assert len(ecowitt_server.sensors) == 52
    assert len(ecowitt_server.stations) == 1


@pytest.mark.asyncio
async def test_server_multi_stations(ecowitt_server, ecowitt_http) -> None:
    """Test server start and multiple stations."""
    sensors = []

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        sensors.append(sensor)

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    resp = await ecowitt_http.post("/", data=EASYWEATHER_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert len(sensors) == 91
    assert len(ecowitt_server.sensors) == 91
    assert len(ecowitt_server.stations) == 2



"""Test for calulcated values from calc.py"""

from aioecowitt import calc

from .const import EASYWEATHER_DATA, GW2000A_DATA


def test_gw2000a_v2():
    """Test Calculated values from GW2000A_V2"""
    values = calc.weather_datapoints(GW2000A_DATA)

    assert values == {
        "PASSKEY": "345544D8EAF42E1B8824A86D8250D5A3",
        "baromabshpa": 944.2,
        "baromabsin": 27.885,
        "baromrelhpa": 944.2,
        "baromrelin": 27.885,
        "dateutc": "2022-07-15 15:26:14",
        "dewpointc": 13.0,
        "dewpointf": 55.3,
        "dewpointinc": 15.7,
        "dewpointinf": 60.2,
        "drain_piezo": 0.0,
        "drain_piezomm": 0.0,
        "erain_piezo": 0.0,
        "erain_piezomm": 0.0,
        "freq": "868M",
        "hrain_piezo": 0.0,
        "hrain_piezomm": 0.0,
        "humidity": 34,
        "humidityin": 53,
        "maxdailygust": 12.53,
        "maxdailygustkmh": 20.2,
        "model": "GW2000A",
        "mrain_piezo": 0.0,
        "mrain_piezomm": 0.0,
        "rrain_piezo": 0.0,
        "rrain_piezomm": 0.0,
        "runtime": "188738",
        "solarradiation": 530.62,
        "solarradiation_lux": 67167.1,
        "stationtype": "GW2000A_V2.1.5",
        "tempc": 30.6,
        "tempf": 87.08,
        "tempfeelsc": 29.7,
        "tempfeelsf": 85.5,
        "tempinc": 26.0,
        "tempinf": 78.8,
        "uv": 4,
        "wh90batt": 2.74,
        "console_batt": 2.5,
        "windchillc": None,
        "windchillf": None,
        "winddir": 289,
        "windgustkmh": 5.0,
        "windgustmph": 3.13,
        "windspeedkmh": 4.7,
        "windspeedmph": 2.91,
        "wrain_piezo": 0.0,
        "wrain_piezomm": 0.0,
        "leaf_batt1": 1.78,
        "leafwetness_ch1": 14,
        "leak_ch1": 0,
        "leakbatt1": 100,
        "soilad1": "120",
        "ws90_ver": "119",
        "ws90cap_volt": 5.4,
        "yrain_piezo": 0.0,
        "yrain_piezomm": 0.0,
    }


def test_easyweather():
    """Test EasyWeather station."""
    values = calc.weather_datapoints(EASYWEATHER_DATA)

    assert values == {
        "PASSKEY": "34271334ED1FADA6D8B988B14267E55D",
        "baromabshpa": 967.8,
        "baromabsin": 28.583,
        "baromrelhpa": 967.8,
        "baromrelin": 28.583,
        "dailyrainin": 0.0,
        "dailyrainmm": 0.0,
        "dateutc": "2020-05-04 14:34:26",
        "dewpointc": 3.9,
        "dewpointf": 39.1,
        "dewpointinc": 8.6,
        "dewpointinf": 47.4,
        "eventrainin": 0.0,
        "eventrainmm": 0.0,
        "freq": "915M",
        "humidity": 32,
        "humidityin": 33,
        "model": "HP3500_V1.6.2",
        "monthlyrainin": 0.079,
        "monthlyrainmm": 2.0,
        "rainratein": 0.0,
        "rainratemm": 0.0,
        "solarradiation": 205.52,
        "solarradiation_lux": 26015.2,
        "stationtype": "EasyWeatherV1.4.9",
        "tempc": 21.3,
        "tempf": 70.3,
        "tempfeelsc": 21.3,
        "tempfeelsf": 70.3,
        "tempinc": 26.1,
        "tempinf": 79.0,
        "uv": 2,
        "weeklyrainin": 0.079,
        "weeklyrainmm": 2.0,
        "windchillc": None,
        "windchillf": None,
        "winddir": 200,
        "windgustkmh": 1.8,
        "windgustmph": 1.1,
        "windspeedkmh": 0.6,
        "windspeedmph": 0.4,
        "yearlyrainin": 0.079,
        "yearlyrainmm": 2.0,
    }



"""Test ecowitt sensor module."""

import pytest

from aioecowitt import server

from .const import GW2000A_V3_DATA
from aioecowitt import sensor


def test_update_listener() -> None:
    """Test on change get updates from callback."""
    ecowit_sensor = sensor.EcoWittSensor(
        "test", "test", sensor.EcoWittSensorTypes.TEMPERATURE_C, "test"
    )

    called = False

    def on_change() -> None:
        """Test callback."""
        nonlocal called
        called = True

    ecowit_sensor.update_cb.append(on_change)

    ecowit_sensor.update_value(10, 0, 0)
    assert called

    called = False
    ecowit_sensor.update_value(10, 0, 0)
    assert not called

    ecowit_sensor.update_value(11, 0, 0)
    assert called


@pytest.mark.asyncio
async def test_heap_field(ecowitt_server, ecowitt_http) -> None:
    """Test handling of heap field."""
    heap_sensor = None

    def on_change(sensor: server.EcoWittSensor) -> None:
        """Test callback."""
        if sensor.key == "heap":
            nonlocal heap_sensor
            heap_sensor = sensor

    ecowitt_server.new_sensor_cb.append(on_change)

    resp = await ecowitt_http.post("/", data=GW2000A_V3_DATA)
    assert resp.status == 200
    text = await resp.text()
    assert text == "OK"

    assert heap_sensor
    assert heap_sensor.value == GW2000A_V3_DATA["heap"]



"""Test for stations from station.py"""

from aioecowitt import station

from .const import EASYWEATHER_DATA, GW2000A_DATA


def test_gw2000a_v2():
    """Test Calculated values from GW2000A_V2"""
    ecowitt_station = station.extract_station(GW2000A_DATA)

    assert "PASSKEY" not in GW2000A_DATA
    assert "model" not in GW2000A_DATA
    assert "frequence" not in GW2000A_DATA
    assert "stationtype" not in GW2000A_DATA
    assert "ws90_ver" not in GW2000A_DATA

    assert ecowitt_station.station == "GW2000A_V2.1.5"
    assert ecowitt_station.key == "345544D8EAF42E1B8824A86D8250D5A3"
    assert ecowitt_station.model == "GW2000A"
    assert ecowitt_station.version == "119"
    assert ecowitt_station.frequence == "868M"


def test_easyweather():
    """Test EasyWeather station."""
    ecowitt_station = station.extract_station(EASYWEATHER_DATA)

    assert "PASSKEY" not in EASYWEATHER_DATA
    assert "model" not in EASYWEATHER_DATA
    assert "frequence" not in EASYWEATHER_DATA
    assert "stationtype" not in EASYWEATHER_DATA

    assert ecowitt_station.station == "EasyWeatherV1.4.9"
    assert ecowitt_station.key == "34271334ED1FADA6D8B988B14267E55D"
    assert ecowitt_station.model == "HP3500_V1.6.2"
    assert ecowitt_station.frequence == "915M"



"""aioEcoWitt API server."""

from __future__ import annotations

import logging
import time
from typing import Callable

from aiohttp import web

from .sensor import EcoWittSensor, SENSOR_MAP
from .station import extract_station, EcoWittStation
from .calc import weather_datapoints

_LOGGER = logging.getLogger(__name__)
_ECOWITT_LISTEN_PORT = 49199


class EcoWittListener:
    """EcoWitt Server API server."""

    def __init__(self, port: int = _ECOWITT_LISTEN_PORT, path: str = None):
        """Initialize EcoWitt Server."""
        # API Constants
        self.port: int = port
        self.path: str = path

        # webserver
        self.server: None | web.Server = None
        self.runner: None | web.ServerRunner = None
        self.site: None | web.TCPSite = None

        # internal data
        self.last_values: dict[str, dict[str, str | None]] = {}
        self.new_sensor_cb: list[Callable[[EcoWittSensor], None]] = []
        self._warned_unknown: set[str] = set()

        # storage
        self.sensors: dict[str, EcoWittSensor] = {}
        self.stations: dict[str, EcoWittStation] = {}

    def _new_sensor_cb(self, sensor: EcoWittSensor) -> None:
        """Internal new sensor callback

        binds to self.new_sensor_cb
        """
        for callback in self.new_sensor_cb:
            callback(sensor)

    def process_data(self, data: dict[str, str | float | int | None]) -> None:
        """Process data from weather station."""
        data = data.copy()
        station = extract_station(data)
        weather_data = weather_datapoints(data)

        # add station to list
        if station.key not in self.stations:
            _LOGGER.debug("Found new station: %s", station.key)
            self.stations[station.key] = station
        else:
            station = self.stations[station.key]

        last_update = time.time()
        last_update_m = time.monotonic()

        for datapoint in weather_data.keys():
            sensor_id = f"{station.key}.{datapoint}"
            sensor = self.sensors.get(sensor_id)
            if sensor is None:
                # we have a new sensor
                if datapoint not in SENSOR_MAP:
                    if datapoint not in self._warned_unknown:
                        self._warned_unknown.add(datapoint)
                        _LOGGER.warning(
                            "Unhandled sensor type %s value %s",
                            datapoint,
                            weather_data[datapoint],
                        )
                    continue
                metadata = SENSOR_MAP[datapoint]

                sensor = EcoWittSensor(
                    metadata.name, datapoint, metadata.stype, station
                )
                self.sensors[sensor_id] = sensor
                try:
                    self._new_sensor_cb(sensor)
                except Exception as err:  # pylint: disable=broad-except
                    _LOGGER.warning("EcoWitt new sensor callback error: %s", err)

            try:
                sensor.update_value(weather_data[datapoint], last_update, last_update_m)
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.warning("Sensor update error: %s", err)

    async def handler(self, request: web.BaseRequest) -> web.Response:
        """AIOHTTP handler for the API."""
        if request.method != "POST":
            return web.Response(status=405)
        if self.path is not None and request.path != self.path:
            return web.Response(status=404)
        data = await request.post()

        # data is not a dict, it's a MultiDict
        self.last_values[data["PASSKEY"]] = data.copy()
        self.last_values[data["PASSKEY"]].pop("PASSKEY")
        self.process_data(data)

        return web.Response(text="OK")

    async def start(self) -> None:
        """Listen and process."""

        self.server = web.Server(self.handler)
        self.runner = web.ServerRunner(self.server)
        await self.runner.setup()
        self.site = web.TCPSite(self.runner, port=self.port)
        await self.site.start()

    async def stop(self) -> None:
        """Stop listening."""
        await self.site.stop()



"""
A bone-simple fake client used to test the hass integration
"""

import http.client
import sys
import urllib.parse

MY_PASSKEY = "34271334ED1FADA6D8B988B14267E55D"
# MY_PASSKEY = '35271334ED1FADA7D8B988B22222E22D'

paramset_a = {
    "PASSKEY": MY_PASSKEY,
    "stationtype": "EasyWeatherV1.4.9",
    "dateutc": "2020-11-13+17:10:24",
    "tempinf": 76.8,
    "humidityin": 22,
    "baromrelin": 28.760,
    "baromabsin": 28.760,
    "tempf": 65.8,
    "humidity": 24,
    "winddir": 319,
    "windspeedmph": 0.9,
    "windgustmph": 1.1,
    "rainratein": 0.000,
    "eventrainin": 0.000,
    "dailyrainin": 0.000,
    "weeklyrainin": 0.024,
    "monthlyrainin": 0.028,
    "yearlyrainin": 0.843,
    "solarradiation": 375.53,
    "uv": 3,
    "pm25_ch1": 8.0,
    "pm25_avg_24h_ch1": 5.2,
    "freq": "915M",
    "model": "HP3500_V1.6.2",
    "leak_ch1": 0,
    "leakbatt1": 5,
}

paramset_b = {
    "PASSKEY": MY_PASSKEY,
    "stationtype": "EasyWeatherV1.5.4",
    "dateutc": "2020-11-16+15:30:24",
    "tempinf": 68.7,
    "humidityin": 52,
    "baromrelin": 29.785,
    "baromabsin": 29.785,
    "tempf": 46.4,
    "humidity": 94,
    "winddir": 260,
    "winddir_avg10m": 260,
    "windspeedmph": 0.0,
    "windspdmph_avg10m": 0.0,
    "windgustmph": 0.0,
    "maxdailygust": 6.9,
    "rainratein": 0.000,
    "eventrainin": 0.118,
    "hourlyrainin": 0.000,
    "dailyrainin": 0.118,
    "weeklyrainin": 0.118,
    "monthlyrainin": 0.378,
    "yearlyrainin": 6.268,
    "solarradiation": 0.00,
    "uv": 0,
    "soilmoisture1": 0,
    "soilad1": 120,
    "wh65batt": 1,
    "wh25batt": 0,
    "console_batt": 2.5,
    "soilbatt1": 1.5,
    "leak_ch1": 0,
    "leakbatt1": 5,
    "leak_ch2": 1,
    "leakbatt2": 3,
    "tf_co2": 56.7,
    "humi_co2": 72,
    "pm25_co2": 24.7,
    "pm25_24h_co2": 29.4,
    "pm10_co2": 24.7,
    "pm10_24h_co2": 29.9,
    "co2": 455,
    "co2_24h": 464,
    "co2_batt": 6,
    "tf_ch1": 71.1,
    "tf_batt1": 1.40,
    "freq": "868M",
    "model": "HP1000SE-PRO_Pro_V1.6.0",
}


def usage():
    print("Usage: {0} host port".format(sys.argv[0]))


if __name__ == "__main__":
    if len(sys.argv) < 3:
        usage()
        exit(1)

    host = sys.argv[1]
    port = sys.argv[2]

    # add a sensor
    if len(sys.argv) > 3 and sys.argv[3] == "add":
        paramset_b["humidity2"] = 21

    print("Connecting to host {0} on port {0}".format(host, port))
    conn = http.client.HTTPConnection(host, port)
    headers = {"Content-type": "application/x-www-form-urlencoded"}
    params = urllib.parse.urlencode(paramset_b)
    print(params)
    conn.request("POST", "", params, headers)
    response = conn.getresponse()
    print(response.status, response.reason)
    conn.close()



"""Run local Test server."""

from __future__ import annotations

import asyncio
import sys

from aioecowitt import EcoWittListener, EcoWittSensor


def usage():
    """Print usage of the CLI."""
    print(f"Usage: {sys.argv[0]} port")


async def my_handler(sensor: EcoWittSensor) -> None:
    """Callback handler for printing data."""
    print("In my handler")
    print(f"{str(sensor)}")


async def run_server(ecowitt_ws: EcoWittListener) -> None:
    """Run server in endless mode."""
    await ecowitt_ws.start()
    while True:
        await asyncio.sleep(100000)


def main() -> None:
    """Run main."""
    if len(sys.argv) < 2:
        usage()
        sys.exit(1)

    print(f"Firing up webserver to listen on port {sys.argv[1]}")
    ecowitt_server = EcoWittListener(port=sys.argv[1])

    ecowitt_server.new_sensor_cb.append(my_handler)
    try:
        asyncio.run(run_server(ecowitt_server))
    except Exception as err:  # pylint: disable=broad-except
        print(str(err))
    print("Exiting")


if __name__ == "__main__":
    main()
